--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  AUTHOR       | Xiaojun Wang                         --
--  -------------+------------------------------------  --
--  DATE         | 27 February 2008                     --
--  -------------+------------------------------------  --
--  REVISED BY   | Jainik Kathiara                      --
--  -------------+------------------------------------  --
--  DATE         | 21 Sept. 2010                        --
--======================================================--

--**********************************************************************************--
--                                                                                  --
--	Copyright (C) 2010		                                            --
--                                                                                  --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                                  --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                                  --
--	You should have received a copy of the GNU General Public License           --
--	along with this program.  If not, see<http://www.gnu.org/licenses/>.        --
--                                                                           	    --
--**********************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

library fp_lib;
use fp_lib.float_pkg.all;

----------------------------------------------------------
--               Floating Point Divider                 --
----------------------------------------------------------
entity divider is
	generic
	(
		bits							: integer
	);
	port
	(
		--inputs
		CLK							: in	std_logic;
		RESET							: in	std_logic;
		STALL							: in	std_logic;
		X								: in	std_logic_vector(2*bits-1 downto 0);
		Y								: in	std_logic_vector(2*bits-1 downto 0);
		--outputs
		RESULT						: out	std_logic_vector(4*bits+7 downto 0)
	);	
end entity;


architecture divider_arch of divider is

	component div_reciprocal_table is
	port 
	(
		clka							: IN std_logic;
		addra							: IN std_logic_VECTOR(bits-1 downto 0);
		ena							: IN std_logic;
		ssra							: IN std_logic;
		douta							: OUT std_logic_VECTOR(2*bits+3 downto 0)
	);
	end component;
	
	component div_multiplier_1 is
	port 
	(
		clk							: IN std_logic;
		a								: IN std_logic_VECTOR(2*bits-1 downto 0);
		b								: IN std_logic_VECTOR(2*bits-1 downto 0);
		ce								: IN std_logic;
		sclr							: IN std_logic;
		p								: OUT std_logic_VECTOR(4*bits-1 downto 0)
	);
	end component;

	component div_multiplier_2 is
	port 
	(
		clk							: IN std_logic;
		a								: IN std_logic_VECTOR(2*bits+3 downto 0);
		b								: IN std_logic_VECTOR(2*bits+3 downto 0);
		ce								: IN std_logic;
		sclr							: IN std_logic;
		p								: OUT std_logic_VECTOR(4*bits+7 downto 0)
	);
	end component;

	--======================================================--
	--                      SIGNALS                         --
	--======================================================--
	signal enable 					: std_logic;
	
	signal x_s 						: std_logic_vector(2*bits-1 downto 0);

	signal yh 						: std_logic_vector(2*bits-1 downto 0);
	signal yl 						: std_logic_vector(2*bits-1 downto 0);

	signal y_sub 					: std_logic_vector(2*bits-1 downto 0);
	signal y_sub_s 				: std_logic_vector(2*bits-1 downto 0);

	signal yhl 						: std_logic_vector(bits-1 downto 0);
	
	signal mul 						: std_logic_vector(4*bits-1 downto 0);
	
	signal invert					: std_logic_vector(2*bits+3 downto 0);
	signal invert_delayed		: std_logic_vector(2*bits+3 downto 0);
	
	signal mul2 					: std_logic_vector(4*bits+7 downto 0);
		
begin
----------------------------------------------------------
--               Component Instantiation                --
----------------------------------------------------------
	enable <= not STALL;
	
	BRAM: div_reciprocal_table
	port map
	(
		clka 						=> CLK,
		ssra 						=> RESET,
		ena  						=> enable,
		addra						=> yhl,
		douta						=> invert
	);	
		
	INVERT_DELAY: bus_delay_block generic map (2*bits+4,DIV_MULTIPLIER_1_DELAY) port map (CLK,RESET,STALL,invert,invert_delayed);

	MUL_1 : div_multiplier_1
	port map
	(
		clk						=> CLK,
		ce							=> enable,
		sclr						=> RESET,
		a							=> x_s,
		b							=> y_sub_s,
		p							=> mul
	);
	
	MUL_2 : div_multiplier_2
	port map
	(
		clk 						=> CLK,
		ce							=> enable,
		sclr						=> RESET,
		a 							=> mul(4*bits-1 downto 2*bits-4),
		b 							=> invert_delayed,
		p 							=> mul2
	);
	
	RESULT <= mul2;
	----------------------------------------------------------
	--              Asynchronous Assignments                --
	----------------------------------------------------------
	yhl <= y(2*bits-2 downto bits-1);
		
	yh(2*bits-1 downto bits-1) <= y(2*bits-1 downto bits-1);
	yh(bits-2 downto 0) <= (others=>'0');
	
	yl(2*bits-1 downto bits-1) <= (others=>'0');
	yl(bits-2 downto 0) <= y(bits-2 downto 0);
	
	DIVISOR_SUB: parameterized_subtractor generic map (2*bits) port map (yh,yl,y_sub);
	
	----------------------------------------------------------
	--               Synchronous Assignments                --
	----------------------------------------------------------
	synchro: process (CLK,RESET,STALL)
	begin
		if(RESET = '1') then
			x_s 			<= (others => '0');
			y_sub_s 		<= (others => '0');
		elsif(rising_edge(CLK) and STALL = '0') then
			x_s 			<= X;
			y_sub_s 		<= y_sub;
		end if;--CLK
	end process;--synchro

end divider_arch; --end of architecture
