--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  AUTHOR       | Pavle Belanovic                      --
--  -------------+------------------------------------  --
--  DATE         | 20 June 2002                         --
--  -------------+------------------------------------  --
--  REVISED BY   | Haiqian Yu                           --
--  -------------+------------------------------------  --
--  DATE         | 18 Jan. 2003                         --
--  -------------+------------------------------------  --
--  REVISED BY   | Jainik Kathiara                      --
--  -------------+------------------------------------  --
--  DATE         | 21 Sept. 2010                        --
--======================================================--

--**********************************************************************************--
--                                                                                  --
--	Copyright (C) 2010		                                            --
--                                                                                  --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                                  --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                                  --
--	You should have received a copy of the GNU General Public License           --
--	along with this program.  If not, see<http://www.gnu.org/licenses/>.        --
--                                                                           	    --
--**********************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_signed.all;

-- float
library fp_lib;
use fp_lib.float_pkg.all;

----------------------------------------------------------
--       Fixed point to floating point conversion       --
----------------------------------------------------------
entity fix2float is
	generic
	(
		fix_bits						: integer := 32;
		exp_bits						: integer := 8;
		man_bits						: integer := 23
	);
	port
	(
		--inputs
		CLK							: in	std_logic;
		RESET							: in	std_logic;
		READY							: in	std_logic;
		STALL							: in	std_logic;
		FIXED							: in	std_logic_vector(fix_bits-1 downto 0);
		ROUND							: in	std_logic;
		EXCEPTION_IN				: in	std_logic;
		--outputs
		FLOAT							: out	std_logic_vector(exp_bits+man_bits downto 0)	:=	(others=>'0');
		EXCEPTION_OUT				: out	std_logic := '0';
		DONE							: out	std_logic := '0'
	);
end fix2float;

----------------------------------------------------------
--       Fixed point to floating point conversion       --
----------------------------------------------------------

architecture fix2float_arch of fix2float is
--======================================================--
--                     FUNCTIONS                        --
--======================================================--
	--function to calculate ceil(log2(n))
	function ceil_log2 (input: integer) return integer is
		variable max		:	integer :=1; -- exp=0 => max=2^0=1
		variable exp		:	integer :=0;
	begin
		log_loop: while max < input loop
			exp	:=	exp+1;	-- increment exponent
			max	:=	max*2;	-- max = 2^exp
		end loop log_loop;
		return exp; -- exponent that gives (2^exp) >= input
	end ceil_log2;

	function maxval (arg1, arg2 : integer) return integer is
	  variable result: integer;
	begin
	  if arg1>arg2 then
	    result:=arg1;
	  else
	    result:=arg2;
	  end if;
	  return result;
	end maxval;
	  
--======================================================--
--                     CONSTANTS                        --
--======================================================--
	constant	bias					: integer := (2**(exp_bits-1))-1;
	constant	shift_bits			: integer := ceil_log2(fix_bits);

	-- modified by xjwang
	constant fix_man_bits 		: integer := maxval(fix_bits,man_bits+1);

--======================================================--
--                      SIGNALS                         --
--======================================================--
	signal	rdy1					:	std_logic := '0';
	signal	rdy2					:	std_logic := '0';
	signal	rdy3					:	std_logic := '0';
	signal	rnd1					:	std_logic := '0';
	signal	rnd2					:	std_logic := '0';
	signal	exc1					:	std_logic := '0';
	signal	exc2					:	std_logic := '0';
	signal	exc3					:	std_logic := '0';
	signal	dummy_exc			:	std_logic := '0';
	signal	cout2					:	std_logic := '0';
	signal	cout3					:	std_logic := '0';
	signal	all_zero				:	std_logic := '0';
	signal	n_all_zero			:	std_logic := '0';
	signal	all_zero1			:	std_logic := '0';
	signal	all_zero2			:	std_logic := '0';
	signal	all_zero3			:	std_logic := '0';
	signal	exc_out3				:	std_logic := '0';
	signal	ctrl					:	std_logic := '0';
	signal	m						:	std_logic_vector(shift_bits-1 downto 0) 		:=	(others=>'0');
	signal	m1						:	std_logic_vector(shift_bits-1 downto 0) 		:=	(others=>'0');
	signal	m1_wide				:	std_logic_vector(exp_bits-1 downto 0) 	 		:=	(others=>'0');
	signal	man1					:	std_logic_vector(fix_bits-1 downto 0)	 		:=	(others=>'0');
	signal	fixed1				:	std_logic_vector(fix_bits-1 downto 0)	 		:=	(others=>'0');
	signal	man2					:	std_logic_vector(fix_man_bits-1 downto 0)		:=	(others=>'0');
	signal	man_out2				:	std_logic_vector(fix_man_bits-1 downto 0)		:=	(others=>'0');
	signal	exp1					:	std_logic_vector(exp_bits-1 downto 0)	 		:=	(others=>'0');
	signal	exp2					:	std_logic_vector(exp_bits-1 downto 0)	 		:=	(others=>'0');
	signal	float2				:	std_logic_vector(exp_bits+man_bits downto 0)	:=	(others=>'0');
	signal	float3				:	std_logic_vector(exp_bits+man_bits downto 0)	:=	(others=>'0');
	signal	float_out			:	std_logic_vector(exp_bits+man_bits downto 0)	:=	(others=>'0');
	signal	zero_float			:	std_logic_vector(man_bits+exp_bits downto 0)	:=	(others=>'0');
	signal	zero_man				:	std_logic_vector(fix_man_bits-1 downto 0)		:=	(others=>'0');
	signal	const					:	std_logic_vector(exp_bits-1 downto 0)			:=	(others=>'0');

begin

--======================================================--
--             COMPONENT INSTANTIATIONS                 --
--======================================================--
	
	--priority encoder
	pri_enc	: parameterized_priority_encoder
	generic map
	(
		man_bits						=>	fix_bits,
		shift_bits					=>	shift_bits
	)
	port map
	(
		MAN_IN						=>	FIXED,
		SHIFT							=>	m,
		EXCEPTION_OUT				=>	dummy_exc
	);
	
	--subtractor
	sub	: parameterized_subtractor
	generic map
	(
		bits							=>	exp_bits
	)
	port map
	(
		A								=>	std_logic_vector(const),
		B								=>	m1_wide,
		O								=>	exp1
	);
	
	--shifter
	var_shf	: parameterized_variable_shifter
	generic map
	(
		bits							=>	fix_bits,
		select_bits					=>	shift_bits,
		direction					=>	'1'
	)
	port map
	(
		I								=>	fixed1,
		S								=>	m1,
		CLEAR							=>	'0',
		O								=>	man1
	);
	
	--adder
	add	: parameterized_adder
	generic map
	(
		bits							=>	fix_man_bits
	)
	port map
	(
		A								=>	man2,
		B								=>	zero_man,
		CIN							=>	'0',
		S								=>	man_out2,
		COUT							=>	cout2
	);
	
	--output mux
	mux	: parameterized_mux
	generic map
	(
		bits							=>	exp_bits+man_bits+1
	)
	port map
	(
		A								=>	zero_float,
		B								=>	float3,
		S								=>	ctrl,
		O								=>	float_out
	);
	--OR gate
	or_gate	: parameterized_or_gate
	generic map
	(
		bits							=>	fix_bits
	)
	port map
	(
		A								=>	FIXED,
		O								=>	n_all_zero
	);

--======================================================--
--            ASYNCHRONOUS SIGNAL ASSIGNMENTS           --
--======================================================--
	all_zero																	<=	NOT(n_all_zero);
	
-- modified by xjwang
--	zero_man																	<=	(others=>'0');
	zero_man(fix_man_bits-1 downto fix_man_bits-fix_bits+2)	<= (others=>'0'); 
	zero_man(fix_man_bits-fix_bits+1)								<= rnd2;
	zero_man(fix_man_bits-fix_bits downto 0) 						<= (others=>'0');
	zero_float																<=	(others=>'0');
	const																		<=	conv_std_logic_vector(fix_bits+bias-1,exp_bits);
	m1_wide(exp_bits-1 downto shift_bits)							<=	(others=>'0');
	m1_wide(shift_bits-1 downto 0)									<=	m1;
	float2(exp_bits+man_bits)											<=	'0'; --all unsigned numbers positive
	float2(exp_bits+man_bits-1 downto man_bits)					<=	exp2;

-- modified by xjwang
--	float2(man_bits-1 downto 0)										<=	man_out2(man_bits downto 1); --truncate rouding bit
	float2(man_bits-1 downto 0)										<=	man_out2(fix_man_bits-1 downto fix_man_bits-man_bits) when fix_bits-1 >= man_bits else man_out2(fix_man_bits-1 downto 1);
	
	--ORs
	exc_out3																	<=	cout3 OR exc3;
	ctrl																		<=	exc_out3 OR all_zero3;

--======================================================--
--             SYNCHRONOUS SIGNAL ASSIGNMENTS           --
--======================================================--
	synchronous: process (CLK,RESET,STALL)
	begin
		if (RESET = '1') then
			--first clock cycle
			rdy1					<=	'0';
			rnd1					<=	'0';
			exc1					<=	'0';
			all_zero1			<=	'0';
			m1						<=	(others => '0');
			fixed1				<=	(others => '0');

			--second clock cycle
			rdy2					<=	'0';
			rnd2					<=	'0';
			exp2					<=	(others => '0');
			
			-- modified by xjwang
			--	man2				<=	man1(fix_bits-2 downto fix_bits-man_bits-2);
			man2(fix_man_bits-1 downto fix_man_bits-fix_bits+1)<= (others => '0'); -- remove hidden 1
			man2(fix_man_bits-fix_bits downto 0) 					<= (others => '0');
			exc2					<=	'0';
			all_zero2			<=	'0';

			--third clock cycle
			rdy3					<=	'0';
			exc3					<=	'0';
			float3				<=	(others => '0');
			cout3					<=	'0';
			all_zero3			<=	'0';

			--fourth clock cycle
			DONE					<=	'0';
			FLOAT					<=	(others => '0');
			EXCEPTION_OUT		<=	'0';		
		elsif(rising_edge(CLK) and STALL = '0') then
			--first clock cycle
			rdy1					<=	READY;
			rnd1					<=	ROUND;
			exc1					<=	EXCEPTION_IN;
			all_zero1			<=	all_zero;
			m1						<=	m;
			fixed1				<=	FIXED;

			--second clock cycle
			rdy2					<=	rdy1;
			rnd2					<=	rnd1;
			exp2					<=	exp1;
			
			-- modified by xjwang
			--	man2				<=	man1(fix_bits-2 downto fix_bits-man_bits-2);
			man2(fix_man_bits-1 downto fix_man_bits-fix_bits+1)<= man1(fix_bits-2 downto 0); -- remove hidden 1
			man2(fix_man_bits-fix_bits downto 0) 					<= (others=>'0');
			exc2					<=	exc1;
			all_zero2			<=	all_zero1;

			--third clock cycle
			rdy3					<=	rdy2;
			exc3					<=	exc2;
			float3				<=	float2;
			cout3					<=	cout2;
			all_zero3			<=	all_zero2;

			--fourth clock cycle
			DONE					<=	rdy3;
			FLOAT					<=	float_out;
			EXCEPTION_OUT		<=	exc_out3;
		end if;
	end process;
end fix2float_arch;
