--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  AUTHOR       | Jainik Kathiara                      --
--  -------------+------------------------------------  --
--  DATE         | 21 Sept 2010                         --
--======================================================--

--**********************************************************************************--
--                                                                                  --
--	Copyright (C) 2010		                                            --
--                                                                                  --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                                  --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                                  --
--	You should have received a copy of the GNU General Public License           --
--	along with this program.  If not, see<http://www.gnu.org/licenses/>.        --
--                                                                           	    --
--**********************************************************************************--

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

library fp_lib;
use fp_lib.float_pkg.all;

entity fp_compare is
	Generic
	(
		exp_bits			: integer :=8;
		man_bits			: integer :=23
	);
	Port 
	( 
		CLK				: in	STD_LOGIC;
      RESET 			: in	STD_LOGIC;
		READY 			: in	STD_LOGIC;
		STALL				: in	STD_LOGIC;
		EXCEPTION_IN	: in	STD_LOGIC;
		OP1 				: in	STD_LOGIC_VECTOR(exp_bits+man_bits downto 0);
		OP2				: in	STD_LOGIC_VECTOR(exp_bits+man_bits downto 0);
		COMPARE_DONE 	: out	STD_LOGIC;
		COMPARE_VALUE	: out	STD_LOGIC_VECTOR(3 downto 0);
		EXCEPTION_OUT	: out STD_LOGIC
	);
end fp_compare;

architecture Behavioral of fp_compare is

	signal large			: STD_LOGIC_VECTOR(exp_bits+man_bits downto 0);
	signal small			: STD_LOGIC_VECTOR(exp_bits+man_bits downto 0);
	
	signal rdy1				: STD_LOGIC;
	signal rdy2				: STD_LOGIC;
	
	signal large_select	: STD_LOGIC;
	signal small_select	: STD_LOGIC;
	signal flip			 	: STD_LOGIC;
	signal flip_1		 	: STD_LOGIC;
	signal flip_2		 	: STD_LOGIC;
	
	signal a_gt_b			: STD_LOGIC;
	signal a_lt_b			: STD_LOGIC;
	signal a_eq_b			: STD_LOGIC;

	signal exp_a_gt_b		: STD_LOGIC;
	signal exp_a_lt_b		: STD_LOGIC;
	signal exp_a_eq_b		: STD_LOGIC;
	
	signal man_a_gt_b		: STD_LOGIC;
	signal man_a_lt_b		: STD_LOGIC;
	signal man_a_eq_b		: STD_LOGIC;	
	
	signal s1_1				: STD_LOGIC;
	signal e1_1				: STD_LOGIC_VECTOR(exp_bits-1 downto 0);
	signal f1_1				: STD_LOGIC_VECTOR(man_bits-1 downto 0);
	signal s2_1				: STD_LOGIC;
	signal e2_1				: STD_LOGIC_VECTOR(exp_bits-1 downto 0);
	signal f2_1				: STD_LOGIC_VECTOR(man_bits-1 downto 0);
	
	signal s1_2				: STD_LOGIC;
	signal f1_2				: STD_LOGIC_VECTOR(man_bits downto 0);
	signal s2_2				: STD_LOGIC;
	signal f2_2				: STD_LOGIC_VECTOR(man_bits downto 0);
	
begin
	
	exponent_compare: parameterized_comparator
	generic map
	(
		bits			=> exp_bits
	)
	port map
	(
		A				=> OP1(exp_bits+man_bits-1 downto man_bits),
		B				=> OP2(exp_bits+man_bits-1 downto man_bits),
		
		A_GT_B		=> exp_a_gt_b,
		A_EQ_B		=> exp_a_eq_b,
		A_LT_B		=> exp_a_lt_b
	);
	
	large_select <= exp_a_gt_b or exp_a_eq_b;
	small_select <= exp_a_lt_b;
	flip <= exp_a_lt_b;
	
	large_mux: parameterized_mux
	generic map
	(
		bits		=> exp_bits+man_bits+1
	)
	port map
	(
		A			=> OP1,
		B			=> OP2,
		S			=> large_select,
		O			=> large
	);

	small_mux: parameterized_mux
	generic map
	(
		bits		=> exp_bits+man_bits+1
	)
	port map
	(
		A			=> OP1,
		B			=> OP2,
		S			=> small_select,
		O			=> small
	);
	
	compare_shift_adjust: shift_adjust 
	generic map
	(
		exp_bits		=>	exp_bits,
		man_bits		=>	man_bits
	)
	port map
	(
		--inputs
		CLK			=>	CLK,
		RESET			=> RESET,
		STALL			=> STALL,
		READY			=>	rdy1,
		E1				=>	e1_1,
		E2				=>	e2_1,
		F2				=>	f2_1,
		--outputs
		E_OUT			=>	open,
		F_OUT			=>	f2_2,
		DONE			=>	rdy2
	);	 

	mantissa_compare: parameterized_comparator
	generic map
	(
		bits		=> man_bits + 1
	)
	port map
	(
		A			=>	f1_2,
		B			=>	f2_2,
		
		A_GT_B	=> man_a_gt_b,
		A_LT_B	=>	man_a_lt_b,
		A_EQ_B	=>	man_a_eq_b
	);

	--permanent connections (signal renaming)	
	a_gt_b <= man_a_lt_b when (flip_2 = '1') else man_a_gt_b;
	a_eq_b <= man_a_eq_b;
	a_lt_b <= man_a_gt_b when (flip_2 = '1') else man_a_lt_b;

	COMPARE_DONE	  <= rdy2;
	
	COMPARE_VALUE(3) <= '1' when ((s1_2 = '0' and s2_2 = '0' and a_lt_b = '1') or
								  (s1_2 = '0' and s2_2 = '1' ) or
								  (s1_2 = '1' and s2_2 = '1' and a_gt_b = '1')) else '0';
	COMPARE_VALUE(2) <= '1' when ((s1_2 = '0' and s2_2 = '0' and a_gt_b = '1') or
								  (s1_2 = '1' and s2_2 = '0' ) or
								  (s1_2 = '1' and s2_2 = '1' and a_lt_b = '1')) else '0';
	COMPARE_VALUE(1) <= '1' when ((s1_2 = '0' and s2_2 = '0' and a_eq_b = '1') or
								  (s1_2 = '1' and s2_2 = '1' and a_eq_b = '1')) else '0';
	COMPARE_VALUE(0) <= '0';
	EXCEPTION_OUT	  <= '0';
	
	----- Synchrounous Signals -----
	process (CLK,RESET,STALL) is
	begin
		if (RESET = '1') then
			rdy1	<= '0';
			
			flip_1<= '0';
			flip_2<= '0';
			
			s1_1	<= '0';
			e1_1	<= (others=>'0');
			f1_1	<= (others=>'0');
			s2_1	<= '0';
			e2_1	<= (others=>'0');
			f2_1	<= (others=>'0');
			
			s1_2	<= '0';
			f1_2	<= (others=>'0');
			s2_2	<= '0';
		elsif (CLK'event and CLK = '1' and STALL = '0') then
			rdy1	<= READY;
			
			flip_1<= flip;
			flip_2<= flip_1;
			
			s1_1	<= large(exp_bits+man_bits);
			e1_1	<= large(exp_bits+man_bits-1 downto man_bits);
			f1_1	<= large(man_bits-1 downto 0);
			s2_1	<= small(exp_bits+man_bits);
			e2_1	<= small(exp_bits+man_bits-1 downto man_bits);
			f2_1	<= small(man_bits-1 downto 0);

			s1_2	<= s1_1;
			f1_2	<= (f1_1 & '0');
			s2_2	<= s2_1;
		end if;
	end process;
			
end Behavioral;

