--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  AUTHOR       | Xiaojun Wang                         --
--  -------------+------------------------------------  --
--  DATE         | 27 February 2008                     --
--  -------------+------------------------------------  --
--  REVISED BY   | Jainik Kathiara                      --
--  -------------+------------------------------------  --
--  DATE         | 21 Sept. 2010                        --
--======================================================--

--**********************************************************************************--
--                                                                                  --
--	Copyright (C) 2010		                                            --
--                                                                                  --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                                  --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                                  --
--	You should have received a copy of the GNU General Public License           --
--	along with this program.  If not, see<http://www.gnu.org/licenses/>.        --
--                                                                           	    --
--**********************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
--use IEEE.std_logic_signed.all;

-- float
library fp_lib;
use fp_lib.float_pkg.all;

----------------------------------------------------------
--               Floating Point Divider                 --
----------------------------------------------------------
entity fp_div is
	generic
	(
		exp_bits								: integer;
		man_bits								: integer
	);
	port
	(
		CLK									: in	std_logic;
		RESET									: in	std_logic;
		STALL									: in	std_logic;
		OP1									: in	std_logic_vector(exp_bits+man_bits downto 0);
		OP2									: in	std_logic_vector(exp_bits+man_bits downto 0);
		READY									: in	std_logic;
		EXCEPTION_IN						: in	std_logic;
		DONE									: out	std_logic;
		RESULT								: out	std_logic_vector(exp_bits+(2*((man_bits+1)/2))+3 downto 0);
		EXCEPTION_OUT						: out	std_logic
	);
end entity;

architecture fp_div_arch of fp_div is

	constant bias		 					: std_logic_vector := CONV_STD_LOGIC_VECTOR((2**(exp_bits-1))-1,exp_bits);
	constant m 								: integer := (man_bits+1)/2;  

	signal s1 								: std_logic;
	signal s2 								: std_logic;
	signal sign	 							: std_logic;
	signal sign_s							: std_logic;
	signal sign_mul1_delayed			: std_logic;
	signal sign_mul2_delayed			: std_logic;
	
	signal e1 								: std_logic_vector(exp_bits-1 downto 0);
	signal e2 								: std_logic_vector(exp_bits-1 downto 0);
	signal e1_ext 							: std_logic_vector(exp_bits downto 0);
	signal e2_ext 							: std_logic_vector(exp_bits downto 0);
	signal exp_unbiased 					: std_logic_vector(exp_bits downto 0);
	signal exp_unbiased_s				: std_logic_vector(exp_bits downto 0);
	signal exp_unbiased_mul1_delayed	: std_logic_vector(exp_bits downto 0);
	signal exp_unbiased_mul2_delayed	: std_logic_vector(exp_bits downto 0);
	signal exp_biased						: std_logic_vector(exp_bits-1 downto 0);
	
	signal f1 								: std_logic_vector(man_bits-1 downto 0);
	signal f2 								: std_logic_vector(man_bits-1 downto 0);
	signal man_divided 					: std_logic_vector(4*m+7 downto 0);
   
	signal rdy								: std_logic;
	signal rdy_s							: std_logic;
	signal rdy_mul1_delayed				: std_logic;
	signal rdy_mul2_delayed				: std_logic;

	signal dividend_exp_all_zero 		: std_logic;
	signal dividend_exp_zero 			: std_logic;
	signal dividend_exp_max 			: std_logic;
	signal divisor_exp_all_zero 		: std_logic;
	signal divisor_exp_zero 			: std_logic;
	signal divisor_exp_max 				: std_logic;

	signal corner 							: std_logic;
	signal corner_s						: std_logic;
	signal corner_mul1_delayed			: std_logic;
	signal corner_mul2_delayed			: std_logic;
	
	signal exp_unbiased_or_zero 		: std_logic;
	signal overflow 						: std_logic;
	signal underflow 						: std_logic;
	signal special 						: std_logic;

	signal exc								: std_logic;
	signal exc_s 							: std_logic;
	signal exc_mul1_delayed				: std_logic;
	signal exc_mul2_delayed				: std_logic;
	
	signal exception						: std_logic;
	signal value 							: std_logic_vector(exp_bits+(2*((man_bits+1)/2))+3 downto 0);
	signal res 								: std_logic_vector(exp_bits+(2*((man_bits+1)/2))+3 downto 0);
	
	signal temp_done						: std_logic;
	signal temp_result					: std_logic_vector(exp_bits+(2*((man_bits+1)/2))+3 downto 0);
	signal temp_exception				: std_logic;
begin
	--------------------------------------------------------------------------------------------
	--														  READY	 													--
	--------------------------------------------------------------------------------------------
	rdy <= READY;
	
	READY_DIV_MULTIPLIER_1_DELAY : delay_block generic map (DIV_MULTIPLIER_1_DELAY) port map (CLK,RESET,STALL,rdy_s,rdy_mul1_delayed);
	READY_DIV_MULTIPLIER_2_DELAY : delay_block generic map (DIV_MULTIPLIER_2_DELAY) port map (CLK,RESET,STALL,rdy_mul1_delayed,rdy_mul2_delayed);
	
	--------------------------------------------------------------------------------------------
	--														 SIGN 														--
	--------------------------------------------------------------------------------------------
	sign <= s1 xor s2;

	SIGN_DIV_MULTIPLIER_1_DELAY : delay_block generic map (DIV_MULTIPLIER_1_DELAY) port map (CLK,RESET,STALL,sign_s,sign_mul1_delayed);
	SIGN_DIV_MULTIPLIER_2_DELAY : delay_block generic map (DIV_MULTIPLIER_2_DELAY) port map (CLK,RESET,STALL,sign_mul1_delayed,sign_mul2_delayed);

	--------------------------------------------------------------------------------------------
	--														EXPONENT 													--
	--------------------------------------------------------------------------------------------
	-- compute the exponent difference e1-e2	
	e1_ext <= '0'&e1; e2_ext <= '0'&e2;
	EXPONENT_UNBIASED_DIFFERENCE: parameterized_subtractor generic map (exp_bits+1) port map (e1_ext,e2_ext,exp_unbiased);

	EXPONENT_DIV_MULTIPLIER_1_DELAY : bus_delay_block generic map (exp_bits+1,DIV_MULTIPLIER_1_DELAY) port map (CLK,RESET,STALL,exp_unbiased_s,exp_unbiased_mul1_delayed);
	EXPONENT_DIV_MULTIPLIER_2_DELAY : bus_delay_block generic map (exp_bits+1,DIV_MULTIPLIER_2_DELAY) port map (CLK,RESET,STALL,exp_unbiased_mul1_delayed,exp_unbiased_mul2_delayed);

	-- add bias to the exponent different after adjust
	EXPONENT_BIASED_DIFFERENCE: parameterized_adder generic map (exp_bits) port map (exp_unbiased_mul2_delayed(exp_bits-1 downto 0),bias,'0',exp_biased,open);

	--------------------------------------------------------------------------------------------
	--														MANTISSA 													--
	--------------------------------------------------------------------------------------------

	mantissa_divider : divider
	generic map
	(
		bits 									=> m
	)
	port map
	(
		CLK 									=> CLK,
		RESET									=> RESET,
		STALL									=> STALL,
	   X 										=> f1,
		Y 										=> f2,
		RESULT 								=> man_divided
	);    

	--------------------------------------------------------------------------------------------
	--														EXCETPTION													--
	--------------------------------------------------------------------------------------------
	exc <= EXCEPTION_IN;

	EXCEPTION_DIV_MULTIPLIER_1_DELAY : delay_block generic map (DIV_MULTIPLIER_1_DELAY) port map (CLK,RESET,STALL,exc_s,exc_mul1_delayed);
	EXCEPTION_DIV_MULTIPLIER_2_DELAY : delay_block generic map (DIV_MULTIPLIER_2_DELAY) port map (CLK,RESET,STALL,exc_mul1_delayed,exc_mul2_delayed);
	
	-- special cases
	-- detect divide by zero, inf, nan, denormal either dividend and divisor
	ALL_ZERO_EXP_DIVIDEND: parameterized_or_gate generic map ( exp_bits ) port map (e1,dividend_exp_all_zero);
	dividend_exp_zero	<= not(dividend_exp_all_zero);

	ALL_ZERO_EXP_DIVISOR	: parameterized_or_gate generic map ( exp_bits ) port map (e2,divisor_exp_all_zero);
	divisor_exp_zero	<= not(divisor_exp_all_zero);

	ALL_ONE_EXP_DIVIDEND	: parameterized_and_gate generic map( exp_bits ) port map (e1,dividend_exp_max);

	ALL_ONE_EXP_DIVISOR	: parameterized_and_gate generic map( exp_bits ) port map (e2,divisor_exp_max);

	corner <= dividend_exp_max or divisor_exp_max or divisor_exp_zero or dividend_exp_zero;

	CORNER_DIV_MULTIPLIER_1_DELAY	: delay_block generic map (DIV_MULTIPLIER_1_DELAY) port map (CLK,RESET,STALL,corner_s,corner_mul1_delayed);
	CORNER_DIV_MULTIPLIER_2_DELAY	: delay_block generic map (DIV_MULTIPLIER_2_DELAY) port map (CLK,RESET,STALL,corner_mul1_delayed,corner_mul2_delayed);

	-- overflow : 01xx..xx
	overflow <= not( exp_unbiased_mul2_delayed(exp_bits) ) and exp_unbiased_mul2_delayed(exp_bits-1);
	-- underflow : 10xx..xx or
	ALL_ZERO_UNDERFLOW	: parameterized_or_gate generic map (exp_bits-1) port map (exp_unbiased_mul2_delayed(exp_bits-1 downto 1),exp_unbiased_or_zero);
	underflow <= ( exp_unbiased_mul2_delayed(exp_bits) and not( exp_unbiased_mul2_delayed(exp_bits-1) ) ) or ( exp_unbiased_mul2_delayed(exp_bits) and exp_unbiased_mul2_delayed(exp_bits-1) and not(exp_unbiased_or_zero) );

	special <= overflow or underflow or exc_mul2_delayed or corner_mul2_delayed;

	--------------------------------------------------------------------------------------------
	--												Asynchronous Assignments										--
	--------------------------------------------------------------------------------------------	
	s1 <= OP1(exp_bits+man_bits);
	s2 <= OP2(exp_bits+man_bits);
	e1 <= OP1(exp_bits+man_bits-1 downto man_bits);
	e2 <= OP2(exp_bits+man_bits-1 downto man_bits);
	f1 <= OP1(man_bits-1 downto 0);
	f2 <= OP2(man_bits-1 downto 0);
	
	-- select between special and non-special
	value		<= sign_mul2_delayed & exp_biased & man_divided(4*m+5 downto 2*m+3);
	res 		<= (others=>'0') when (special='1') else value;
	exception<= rdy_mul2_delayed and special;
   
	--------------------------------------------------------------------------------------------
	--               							Synchronous Assignments                					--
	--------------------------------------------------------------------------------------------
	synchro: process (CLK,RESET,STALL) is
	begin
		if(RESET = '1') then
			rdy_s						<= '0';
			sign_s					<= '0';			
			exp_unbiased_s			<= (others => '0');			
			corner_s					<= '0';
			exc_s						<= '0';
			
			temp_done				<= '0';
			temp_result				<= (others=> '0');
			temp_exception			<= '0';
		elsif(rising_edge(CLK) and STALL = '0') then
			rdy_s						<= rdy;
			sign_s					<= sign;
			exp_unbiased_s			<= exp_unbiased;
			corner_s					<= corner;
			exc_s						<= exc;

			temp_done				<= rdy_mul2_delayed;
			temp_result				<= res;
			temp_exception			<= exception;
		end if;
	end process;

	DONE				<= temp_done;
	RESULT 			<= temp_result;
	EXCEPTION_OUT 	<= temp_exception and temp_done;
	
end fp_div_arch; 
